#!/bin/python3
#%%
import os, re
import numpy as np
import pandas as pd
from tools.std_lib.basic import json_dump
# from tools.mdl_kit_v6 import get_distri_by_col_dict, get_info_value_dict, stat_features_df_by_col, stat_features_info_value_by_col, stat_features_psi_by_col, t_print, pickle_dump, save_and_reset_index
from tools.data_processing.easy_pandas import t_print, pickle_dump, save_and_reset_index, multiprocessing_apply

#%%

def get_qcut_bins(series: pd.Series, q = 10, mode = None):
    series = series[~series.isna()]
    if mode == 'auto' and len(series) > 0:
        uniques = sorted(series.unique().tolist())
        if len(uniques) <= 10:
            bins = uniques
        else:
            cnt_sr = series.value_counts()
            cnt_sr = cnt_sr / cnt_sr.sum()
            bins1 = cnt_sr[cnt_sr >= .05].index.to_list()
            bins2 = cnt_sr[cnt_sr < .05].index.to_list()
            new_series = series[series.isin(bins2)]
            if len(new_series) > 0:
                new_bins = pd.qcut(new_series, q=q, duplicates='drop', retbins=True)[1].tolist()
                bins = sorted(set(bins1) | set(new_bins))
            else:
                bins = sorted(bins1)
    elif len(series) > 0:
        bins = pd.qcut(series, q=q, duplicates='drop', retbins=True)[1]
        bins = bins.tolist()
    elif len(series) == 0:
        bins = []
    bins = sorted(bins)
    return bins


def get_qcut_bins_auto(data_sr: pd.Series):
    return get_qcut_bins(data_sr, mode='auto')


def get_features_bins_dict(data_df: pd.DataFrame, disable_multiprocessing = False):
    t_print('get_features_bins_dict()')
    n1 = data_df.shape[1]
    data_df = data_df.select_dtypes(include='number')
    n2 = data_df.shape[1]
    t_print(f'select_dtypes: n_cols: {n1} - {n1 - n2} = {n2}')
    features_bins_dict = multiprocessing_apply(
            get_qcut_bins_auto, 
            {col: data_df[col] for col in data_df},
            disable_multiprocessing=disable_multiprocessing
            )
    return features_bins_dict


def cal_bins_and_cnts(series: pd.Series, bins = 10, percent=False, ret = 'series'):
    if type(bins) == int and bins > 0:
        bins = get_qcut_bins(series, q = bins)
    bins = [-float('inf')] + bins + [float('inf')]
    cut_sr = pd.cut(series, bins = bins)
    cnt_sr = cut_sr.value_counts().sort_index()
    cnt_sr['Not_a_Number'] = series.isna().sum()
    if percent :
        cnt_sr = cnt_sr / cnt_sr.sum()
    r = cnt_sr
    if ret == 'dict':
        cnt_sr.index = cnt_sr.index.to_series().apply(lambda s: s.__str__())
        cnt_dict = cnt_sr.to_dict()
        r = cnt_dict
    return r



def get_col_distri_df(args):
    data_sr, bins_dict = args
    cnt_sr = cal_bins_and_cnts(data_sr, bins=bins_dict[data_sr.name])
    distri_df = pd.DataFrame({'cnt': cnt_sr, 'ratio': cnt_sr / cnt_sr.sum()})
    return distri_df


def get_distri_by_col_dict(data_df: pd.DataFrame, by_col: str, disable_multiprocessing = False):
    t_print('# get_distri_by_col_dict()')
    features_bins_dict = get_features_bins_dict(data_df=data_df, disable_multiprocessing=disable_multiprocessing)
    items = sorted(data_df[by_col].unique())

    distri_by_col_dict = {}
    for c, item in enumerate(items, start=1):
        print()
        t_print(f'### {c}/{len(items)} item = {item}')
        idx = data_df[by_col] == item
        args_list = [(data_df.loc[idx, col], features_bins_dict) for col in features_bins_dict]
        distri_by_col_dict[item] = multiprocessing_apply(
            func=get_col_distri_df, 
            vars={col: args for col, args in zip(features_bins_dict, args_list)},
            disable_multiprocessing=disable_multiprocessing
        )
        del idx
    t_print('# end of get_distri_by_col_dict()')
    return distri_by_col_dict


#%%


def find_month_str(string: str):
    pattern = '\d{4}-\d{2}'
    return re.findall(pattern, string)


def get_psi_adjacent_cols(psi_df: pd.DataFrame):
    all_months = sum([find_month_str(k) for k in psi_df], start = [])
    all_months = sorted(set(all_months))
    cols = [f'psi_btw_{all_months[k]}_and_{all_months[k+1]}' for k in range(len(all_months)-1)]
    return cols


def get_trending(s: pd.Series):
    df = s.to_frame()
    df['index'] = range(1, len(s)+1)
    result = df.corr(method='spearman').iloc[0, 1]
    return result


def get_stat_cnt_df(stat_cnt_dict: dict, coverage = .02, gini = .02):

    dfs = []
    for key, stat_df in stat_cnt_dict.items():
        stat_df = stat_df.copy()
        stat_df.set_index('col_name', inplace = True)
        stat_cols = ['total', 'coverage', 'gini']
        stat_df = stat_df[stat_cols].add_prefix(f'{key}-')
        dfs.append(stat_df)
        del stat_df
    stat_df = pd.concat(dfs, axis = 1)

    coverage_cols = stat_df.columns.str.endswith('-coverage')
    coverage_cols = stat_df.columns[coverage_cols].to_list()
    by_coverage_to_mdl = (stat_df.loc[:, coverage_cols] >= coverage).all(axis = 1)
    stat_df['mean-coverage'] = stat_df.loc[:, coverage_cols].mean(axis=1)
    stat_df['min-coverage'] = stat_df.loc[:, coverage_cols].min(axis=1)
    stat_df['max-coverage'] = stat_df.loc[:, coverage_cols].max(axis=1)
    stat_df['abs_diff-coverage'] = stat_df['max-coverage'] - stat_df['min-coverage']
    stat_df['rel_diff-coverage'] = stat_df['abs_diff-coverage'] / stat_df['mean-coverage']
    stat_df['trend-coverage'] = stat_df[coverage_cols].apply(get_trending, axis = 1)
    stat_df['by_coverage-to_mdl'] = by_coverage_to_mdl

    gini_cols = stat_df.columns.str.endswith('-gini')
    gini_cols = stat_df.columns[gini_cols].to_list()
    by_gini_to_mdl = (stat_df.loc[:, gini_cols] >= gini).all(axis = 1)
    stat_df['mean-gini'] = stat_df.loc[:, gini_cols].mean(axis=1)
    stat_df['min-gini'] = stat_df.loc[:, gini_cols].min(axis=1)
    stat_df['max-gini'] = stat_df.loc[:, gini_cols].max(axis=1)
    stat_df['abs_diff-gini'] = stat_df['max-gini'] - stat_df['min-gini']
    stat_df['rel_diff-gini'] = stat_df['abs_diff-gini'] / stat_df['mean-gini']
    stat_df['trend-gini'] = stat_df[gini_cols].apply(get_trending, axis = 1)
    stat_df['by_gini-to_mdl'] = by_gini_to_mdl

    stat_df['to_mdl'] = stat_df['by_coverage-to_mdl'] & stat_df['by_gini-to_mdl']

    print('\nmin-coverage describe:\n', stat_df['min-coverage'].describe())
    print('\nmin-gini describe:\n', stat_df['min-gini'].describe())

    stat_df = save_and_reset_index(stat_df)

    return stat_df


def get_stat_distri_df(distri_by_col_dict: dict):
    distri_df = pd.DataFrame({
        kk: pd.Series({k: v.to_json() for k, v in vv.items()}) for kk, vv in distri_by_col_dict.items()
        }).add_prefix('distri_of_')
    distri_df = save_and_reset_index(distri_df, origin_index_name='col_name')
    return distri_df


def fix_psi_df(psi_df: pd.DataFrame) -> pd.DataFrame:
    
    adjacent_cols = get_psi_adjacent_cols(psi_df)
    adjacent_psi_df = psi_df[adjacent_cols]
    psi_df['adjacent_min_psi'] = adjacent_psi_df.min(axis=1)
    psi_df['adjacent_mean_psi'] = adjacent_psi_df.mean(axis=1)
    psi_df['adjacent_max_psi'] = adjacent_psi_df.max(axis=1)
    psi_df['adjacent_std_psi'] = adjacent_psi_df.std(axis=1)
    psi_df['adjacent_rsd_psi'] = psi_df['adjacent_std_psi'] / psi_df['adjacent_mean_psi']

    return psi_df


def stat_series(df_col, mod = 1, total = None, to_mdl = {'coverage': .01, 'gini': .01}):
    if total == None:
        total = len(df_col)
    stat_sr = pd.Series(dtype=object)
    stat_sr['col_name'] = str(df_col.name)
    stat_sr['dtype'] = str(df_col.dtype.name)
    stat_sr['length'] = len(df_col)
    stat_sr['total'] = total
    stat_sr['len_nan'] = df_col.isna().sum()
    stat_sr['len_non_nan'] = stat_sr['length'] - stat_sr['len_nan']
    stat_sr['coverage'] = (stat_sr['length'] - stat_sr['len_nan']) / stat_sr['total']
    stat_sr['sample'] = df_col.loc[~df_col.isna()][:10].to_list()
    if mod == 1:
        stat_sr['nunique'] = df_col.nunique()
        value_counts = df_col.value_counts()
        stat_sr['value_counts'] = value_counts.iloc[:10].to_dict()
        gini = 1 - ((value_counts / value_counts.sum())**2).sum() if len(value_counts) > 0 else float('nan')
        stat_sr['gini'] = gini
        stat_sr['to_mdl'] = False
        if df_col.dtype.name in ['Int64', 'int64', 'Int32', 'int32', 'Float', 'float', 'Float64', 'float64', 'datetime64[ns]']:
            stat_sr['min'] = df_col.min()
            stat_sr['max'] = df_col.max()
            if df_col.dtype.name not in ['datetime64[ns]']:
                stat_sr['mean'] = df_col.mean()
                stat_sr['std'] = df_col.std()
            else:
                stat_sr['mean'] = float('nan')
                stat_sr['std'] = float('nan')
            
            if not pd.isna(stat_sr['min']) and stat_sr['min'] < stat_sr['max'] and df_col.dtype.name not in ['datetime64[ns]']:
                if stat_sr['coverage'] >= to_mdl['coverage'] and stat_sr['gini'] >= to_mdl['gini']:
                    stat_sr['to_mdl'] = True
        else:
            stat_sr['min'] = float('nan')
            stat_sr['max'] = float('nan')
            stat_sr['mean'] = float('nan')
            stat_sr['std'] = float('nan')
    return stat_sr


def stat_dataframe_multiprocessing(df: pd.DataFrame, max_processes = .9, disable_multiprocessing = False):
    cols = [df[col] for col in df]
    stat_df = multiprocessing_apply(
        stat_series, 
        cols, 
        max_processes=max_processes, 
        disable_multiprocessing=disable_multiprocessing
        )
    stat_df = pd.DataFrame(stat_df)
    stat_df.index = pd.Index(range(1, len(stat_df)+1), name='index')
    return stat_df


def stat_features_df_by_col(data_df: pd.DataFrame, by_col: str, max_processes = .9, disable_multiprocessing = False):
    t_print('# stat_features_df_by_col()')
    items = sorted(data_df[by_col].unique().tolist())
    num_cols = data_df.select_dtypes(include='number').columns.to_list()
    stat_dict = {}
    for k, item in enumerate(items, start=1):
        t_print(f'## {k}/{len(items)} item =', item)
        idx = data_df[by_col] == item
        part_df = data_df.loc[idx, num_cols]
        stat_dict[item] = stat_dataframe_multiprocessing(part_df, max_processes = max_processes, disable_multiprocessing=disable_multiprocessing)
        print()
    t_print('items =', items)
    t_print('end of stat_features_df_by_col()')
    return stat_dict


def get_feature_psi(args):
    distri_by_col_dict, a_item, b_item, feature = args
    psi = cal_psi_value(
        distri_by_col_dict[a_item][feature]['ratio'], 
        distri_by_col_dict[b_item][feature]['ratio']
        )
    return psi


def stat_features_psi_by_col(distri_by_col_dict: dict, max_processes = .9, disable_multiprocessing =False):
    t_print('# stat_features_psi_by_col()')
    items = sorted(set(distri_by_col_dict))
    psi_dict = {}

    total = int(len(items) * (len(items) - 1) / 2)
    t_print('total =', total)
    cnt = 0

    for a_item in items:
        for b_item in items:
            if not(a_item < b_item):
                continue
            cnt += 1
            a_month_dict = distri_by_col_dict[a_item]
            b_month_dict = distri_by_col_dict[b_item]
            features = sorted(set(a_month_dict) & set(b_month_dict))
            if not (len(features) == len(a_month_dict) == len(b_month_dict)):
                raise ValueError
            args_list = [(distri_by_col_dict, a_item, b_item, feature) for feature in features]
            print()
            key = f'psi_btw_{a_item}_and_{b_item}'
            t_print(f'## {cnt}/{total} {key}')
            psi_dict[key] = multiprocessing_apply(
                get_feature_psi, 
                {k: v for k, v in zip(features, args_list)}, 
                max_processes=max_processes,
                disable_multiprocessing=disable_multiprocessing
                )

    psi_df = pd.DataFrame(psi_dict)
    psi_copy_df = psi_df.copy()
    psi_df['min_psi'] = psi_copy_df.min(axis=1)
    psi_df['mean_psi'] = psi_copy_df.mean(axis=1)
    psi_df['max_psi'] = psi_copy_df.max(axis=1)
    psi_df['std_psi'] = psi_copy_df.std(axis=1)
    psi_df['rsd_psi'] = psi_df['std_psi'] / psi_df['mean_psi']

    print('\nmax_psi:\n', psi_df['max_psi'].describe())

    psi_df = save_and_reset_index(psi_df, 'col_name')
    t_print('end of stat_features_psi_by_col()')
    return psi_df


def save_monthly_stat_of_cnt_and_psi(
        prefix: str, 
        modeling_path: str,
        data_df: pd.DataFrame, 
        by_col: str, 
        ):

    stat_cnt_dict = stat_features_df_by_col(data_df=data_df, by_col=by_col, disable_multiprocessing=False)

    cnt_stat_df = get_stat_cnt_df(stat_cnt_dict)
    valid_cols = cnt_stat_df['col_name'][cnt_stat_df['to_mdl']].to_list()

    stat_cnt_dir = os.path.join(modeling_path, 'stat/feature/cnt')
    os.makedirs(stat_cnt_dir, exist_ok=True)

    cnt_dict_pickle_name = f'{prefix}-feature_stat-by_month-cnt-pickle'
    save_path = os.path.join(stat_cnt_dir, cnt_dict_pickle_name)
    pickle_dump(stat_cnt_dict, save_path)

    cnt_features_name = f'{prefix}-feature_stat-by_month-cnt.csv'
    save_path = os.path.join(stat_cnt_dir, cnt_features_name)
    cnt_stat_df.to_csv(save_path)

    cnt_features_name = f'{prefix}-valid_features-by_month-cnt.json'
    save_path = os.path.join(stat_cnt_dir, cnt_features_name)
    json_dump(valid_cols, save_path)

    t_print('stat_cnt saved.')

    data_df = data_df[[by_col] + valid_cols]

    distri_by_col_dict = get_distri_by_col_dict(data_df, by_col, disable_multiprocessing=False)
    print('\n\n')
    psi_df = stat_features_psi_by_col(distri_by_col_dict, disable_multiprocessing=True)
    distri_df = get_stat_distri_df(distri_by_col_dict)

    psi_df = fix_psi_df(psi_df)

    stat_psi_dir = os.path.join(modeling_path, 'stat/feature/psi')
    os.makedirs(stat_psi_dir, exist_ok=True)

    pickle_name =f'{prefix}-feature_stat-by_month-distri-pickle'
    save_path = os.path.join(stat_psi_dir, pickle_name)
    pickle_dump(distri_by_col_dict, save_path)

    pickle_name =f'{prefix}-feature_stat-by_month-psi.csv'
    save_path = os.path.join(stat_psi_dir, pickle_name)
    psi_df.to_csv(save_path)

    pickle_name =f'{prefix}-feature_stat-by_month-distri.csv'
    save_path = os.path.join(stat_psi_dir, pickle_name)
    distri_df.to_csv(save_path)

    t_print('stat_psi saved.')

    t_print(f'end of prefix = {prefix}\n\n\n')

    return cnt_stat_df, psi_df #, distri_df


def get_info_value_dict(data_df: pd.DataFrame, by_col: str, disable_multiprocessing = False):
    t_print('# get_info_value_dict()')
    items = sorted(data_df[by_col].unique())
    info_value_dict = {}
    for k, item in enumerate(items, start=1):
        t_print(f'## {k}/{len(items)} item = {item}')
        item_df = data_df.loc[data_df['month']==item, :]
        info_value_dict[item] = get_distri_by_col_dict(item_df, by_col='y', disable_multiprocessing=disable_multiprocessing)
        print()
    return info_value_dict


def cal_psi_value(a_ratio_sr: pd.Series, b_ratio_sr: pd.Series):
    df = pd.DataFrame({
        'a': a_ratio_sr.copy().replace(0, 1e-6),
        'b': b_ratio_sr.copy().replace(0, 1e-6)
    }).fillna(1e-6)
    psi_sr = (df['a'] - df['b']) * np.log(df['a'] / df['b'])
    psi_sum = psi_sr.sum()
    return psi_sum


def get_feature_info_value(args):
    info_value_dict, item, feature = args
    info_value = cal_psi_value(
        info_value_dict[item]['0'][feature]['ratio'], 
        info_value_dict[item]['1'][feature]['ratio']
        )
    return info_value


def stat_features_info_value_by_col(info_value_dict: dict, max_processes=.9, disable_multiprocessing=True):
    t_print('# stat_features_info_value_by_col()')
    items = sorted(set(info_value_dict))
    iv_dict = {}
    for k, item in enumerate(items, start=1):
        pos_dict = info_value_dict[item]['1']
        neg_dict = info_value_dict[item]['0']
        features = sorted(set(pos_dict) & set(neg_dict))
        if not (len(features) == len(pos_dict) == len(neg_dict)):
            raise ValueError
        args_list = [(info_value_dict, item, feature) for feature in features]
        print()
        key = f'info_value_of_{item}'
        t_print(f'# {k}/{len(items)} {key}')
        iv_dict[key] = multiprocessing_apply(
            get_feature_info_value,
            {k: v for k, v in zip(features, args_list)},
            max_processes=max_processes,
            disable_multiprocessing=disable_multiprocessing
        )
    iv_df = pd.DataFrame(iv_dict)
    iv_copy_df = iv_df.copy()
    iv_df['trend_info_value'] = iv_copy_df.apply(get_trending, axis=1)
    iv_df['min_info_value'] = iv_copy_df.min(axis=1)
    iv_df['mean_info_value'] = iv_copy_df.mean(axis=1)
    iv_df['std_info_value'] = iv_copy_df.std(axis=1)
    iv_df['rsd_info_value'] = iv_df['std_info_value'] / iv_df['mean_info_value']
    iv_df['max_info_value'] = iv_copy_df.max(axis=1)
    iv_df['abs_diff_info_value'] = iv_df['max_info_value'] - iv_df['min_info_value']
    iv_df['rel_diff_info_value'] = iv_df['abs_diff_info_value'] / iv_df['mean_info_value']
    iv_df = save_and_reset_index(iv_df, origin_index_name='col_name')
    print('\nmin_info_value:\n', iv_df['min_info_value'].describe())
    return iv_df


def save_monthly_stat_of_info_value(
    prefix: str, 
    modeling_path: str,
    data_df: pd.DataFrame, 
    by_col: str,):
    info_value_dict = get_info_value_dict(data_df=data_df, by_col=by_col)
    info_value_df = stat_features_info_value_by_col(info_value_dict=info_value_dict, disable_multiprocessing=True)

    stat_iv_dir = os.path.join(modeling_path, 'stat/feature/info_value')
    os.makedirs(stat_iv_dir, exist_ok=True)

    pickle_name =f'{prefix}-feature_stat-by_month_by_y-distri-pickle'
    save_path = os.path.join(stat_iv_dir, pickle_name)
    pickle_dump(info_value_dict, save_path)

    pickle_name =f'{prefix}-feature_stat-by_month-info_value.csv'
    save_path = os.path.join(stat_iv_dir, pickle_name)
    info_value_df.to_csv(save_path)

    t_print('stat_info_value saved.')

    print(f'end of prefix = {prefix}\n\n\n')

    return info_value_df


#%%

def get_cnt_features(cnt_path: str, coverage = .02, gini = .02):
    t_print({
        'coverage': coverage,
        'gini': gini
    })
    cnt_df = pd.read_csv(cnt_path, index_col=0)
    idx = (cnt_df['min-coverage'] >= coverage) & (cnt_df['min-gini'] >= gini) # !!!
    n1 = len(cnt_df)
    cnt_df = cnt_df.loc[idx, :]
    n2 = len(cnt_df)
    cnt_features = cnt_df['col_name'].to_list()
    t_print(f'cnt: {n1} - {n1-n2} = {n2}')
    t_print('coverage:', [
        cnt_df['min-coverage'].min(), 
        cnt_df['min-coverage'].max()
        ])
    t_print('gini:', [
        cnt_df['min-gini'].min(), 
        cnt_df['min-gini'].max()
        ])
    del cnt_df
    return cnt_features


def get_psi_features(psi_path: str, psi = .2, adjacent_max_psi = .1):
    t_print({
        'psi_path': psi_path,
        'adjacent_max_psi': adjacent_max_psi
    })
    psi_df = pd.read_csv(psi_path, index_col=0)
    idx = (psi_df['max_psi'] < psi) & (psi_df['adjacent_max_psi'] < adjacent_max_psi) # !!!
    n1 = len(psi_df)
    psi_df = psi_df.loc[idx, :]
    n2 = len(psi_df)
    psi_features = psi_df['col_name'].to_list()
    t_print(f'psi: {n1} - {n1-n2} = {n2}')
    t_print('adjacent_max_psi:', [
        psi_df['adjacent_max_psi'].min(), 
        psi_df['adjacent_max_psi'].max()
        ])
    t_print('psi:', [
        psi_df['max_psi'].min(), 
        psi_df['max_psi'].max()
        ])
    del psi_df
    return psi_features


def get_info_value_features(info_value_path: str, info_value_n = 500, min_info_value = 1e-6, valid_features = None):
    t_print({
        'info_value_n': info_value_n,
        'min_info_value': min_info_value
    })
    info_value_df = pd.read_csv(info_value_path, index_col=0)
    if valid_features is not None:
        info_value_df = info_value_df.loc[info_value_df['col_name'].isin(valid_features), :]
    info_value_df = info_value_df.loc[info_value_df['min_info_value'] > min_info_value, :]
    n1 = len(info_value_df) # !!!
    idx = info_value_df['max_info_value'] - info_value_df['min_info_value'] < .5
    info_value_df = info_value_df.loc[idx, :]
    idx = info_value_df['min_info_value'].nlargest(info_value_n).sort_index().index
    info_value_df = info_value_df.loc[idx, :]
    n2 = len(info_value_df)
    info_value_features = info_value_df['col_name'].to_list()
    t_print(f'info_value: {n1} - {n1-n2} = {n2}')
    t_print('info_value:', [
        info_value_df['min_info_value'].min(),
        info_value_df['max_info_value'].max()
        ])
    del info_value_df
    return info_value_features

